#include <msp/core/maputils.h>
#include <msp/gl/meshbuilder.h>
#include "libr2c2/trackappearance.h"
#include "catalogue.h"
#include "signaltype.h"
#include "tracktype.h"
#include "vehicletype.h"

using namespace std;
using namespace Msp;

namespace R2C2 {

Catalogue3D::Catalogue3D(Catalogue &c):
	catalogue(c)
{
	const list<DataFile::CollectionSource *> &src = catalogue.get_sources();
	for(list<DataFile::CollectionSource *>::const_iterator i=src.begin(); i!=src.end(); ++i)
		add_source(**i);
	catalogue.signal_source_added.connect(sigc::mem_fun(static_cast<DataFile::Collection *>(this), &Catalogue3D::add_source));

	set_srgb_conversion(true);
}

Catalogue3D::~Catalogue3D()
{
	for(map<const ObjectType *, ObjectType3D *>::iterator i=objects.begin(); i!=objects.end(); ++i)
		delete i->second;
}

const ObjectType3D &Catalogue3D::get_3d(const ObjectType &ot)
{
	ObjectMap::iterator i = objects.find(&ot);
	if(i!=objects.end())
		return *i->second;

	ObjectType3D *ot3d = 0;
	if(const TrackType *tt = dynamic_cast<const TrackType *>(&ot))
		ot3d = new TrackType3D(*this, *tt);
	else if(const SignalType *st = dynamic_cast<const SignalType *>(&ot))
		ot3d = new SignalType3D(*this, *st);
	else if(const VehicleType *vt = dynamic_cast<const VehicleType *>(&ot))
		ot3d = new VehicleType3D(*this, *vt);
	else
		throw key_error(&ot);

	objects[&ot] = ot3d;
	return *ot3d;
}

const ObjectType3D &Catalogue3D::get_3d(const ObjectType &ot) const
{
	return *get_item(objects, &ot);
}

const GL::Mesh &Catalogue3D::get_endpoint_mesh(const TrackAppearance &appearance)
{
	EndpointMap::iterator i = endpoint_meshes.find(&appearance);
	if(i!=endpoint_meshes.end())
		return *i->second;

	const Profile &ballast_profile = appearance.get_ballast_profile();
	const Vector &ballast_min = ballast_profile.get_min_coords();
	const Vector &ballast_max = ballast_profile.get_max_coords();

	const Profile &rail_profile = appearance.get_rail_profile();
	const Vector &rail_min = rail_profile.get_min_coords();
	const Vector &rail_max = rail_profile.get_max_coords();

	float gauge = appearance.get_gauge();

	float width = max(max(-ballast_min.x, ballast_max.x)*2, gauge+(rail_max.x-rail_min.x)*2)+0.004;
	float height = ballast_max.y-ballast_min.y+rail_max.y-rail_min.y+0.01;

	GL::Mesh *mesh = new GL::Mesh((GL::NORMAL3, GL::VERTEX3));
	GL::MeshBuilder bld(*mesh);
	bld.normal(1, 0, 0);
	bld.begin(GL::QUADS);
	bld.vertex(0, width/2, 0);
	bld.vertex(0, width/2, height);
	bld.vertex(0, -width/2, height);
	bld.vertex(0, -width/2, 0);
	bld.end();

	endpoint_meshes[&appearance] = mesh;
	return *mesh;
}

const GL::Mesh &Catalogue3D::get_endpoint_mesh(const TrackAppearance &appearance) const
{
	return *get_item(endpoint_meshes, &appearance);
}

} // namespace R2C2
